/*--
	 client.c
	 Protocol on the wire is as follows;

	 4 byte header	- specifies length of data to follow
	 data			- the actual data

	 Client, when running on Windows, will perform 
	         ASCII<-->EBCDIC conversion	
--*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>

#ifdef WIN32
    #include "winsock2.h"
#else
    #include <sys/socket.h>
    #include <netinet/in.h>
    #include <arpa/inet.h>
    #include <netdb.h>
#endif

#ifdef WIN32
/*-- define function mappings --*/
    #define close(x)			closesocket(x)
    #define read(x, y, z)		recv((x), (y), (z), 0)
    #define write(x, y, z)		send((x), (y), (z), 0)

    /*-- Winsock setup stuff --*/
    static WSADATA wskData;
    static int winSockInit;

    long convertAsciiToEbcdic(unsigned char *inBuffer,  long inLength, 
    						  unsigned char *outBuffer, long outLength);
    long convertEbcdicToAscii(unsigned char *inBuffer,  long inLength, 
    						  unsigned char *outBuffer, long outLength);
#endif

char *show_address(struct in_addr *myaddr);

typedef struct _CONNECTION
{
	int					connected;
	int					theSocket;
	char				service[256];
	char				my_inet_addr[256];
	short				thePort;
	struct hostent		*he;
	struct sockaddr_in	serv_addr;
} CONNECTION, *PCONNECTION;

struct servent		*se;

int main(int argc, char **argv)
{

	int				charCount;
	FILE			*fpIn, *fpOut;
	char            inFilename []  = "infile";
	char            outFilename [] = "outfile";

	PCONNECTION		session;
	unsigned char	*txtPtr;
	unsigned char	txt_data[10000];
	unsigned char	*replyPtr;
	unsigned char	*reply_data;

	/*-- 
		 this needs to be set to the host name
		 or IP address of the host you want to
		 connect to 
	--*/
	char            thehost[256];
	strcpy(thehost, argv[1]);
 
    if ((fpIn = fopen(inFilename, "rb")) == NULL) {
	    fprintf(stderr, "FILE ERROR : file 'infile' not found\n");
        return -10;
    }	
	else {
        charCount = fread(txt_data, 1, 10000, fpIn);
	    printf("%d chars read\n", charCount);
		fclose(fpIn);
    }

#ifdef WIN32
    memset(&wskData, 0, sizeof(WSADATA));
    winSockInit=WSAStartup(0x202, &wskData);
    
    if(winSockInit != 0)
    {
        fprintf(stderr, "TCP ERROR : Winsock failed to init\n");
        return -1;
    }
#endif

	session = calloc(sizeof(CONNECTION), 1); 

    strcpy(session->service, "8888");
	session->he = gethostbyname(thehost);
	strcpy(session->my_inet_addr,
			show_address((struct in_addr *)session->he->h_addr));

#ifdef __ILEC400__
	se = getservbyname(session->service, "TCP");
#else
	se = getservbyname(session->service, "tcp");
#endif
	if(se)
	{
		session->thePort = htons(se->s_port);
	}
	else
	{
		if ((session->thePort = atoi(session->service)) == 0)
		{
			fprintf(stderr, "TCP ERROR : Cannot find service\n");
			return -1;
		}
	}

	if (session->thePort == 0)
	{
		fprintf(stderr, "TCP ERROR : No [service] given\n");
		return -1;
	}

	memset(&session->serv_addr, 0, sizeof(session->serv_addr));
	session->serv_addr.sin_family      = AF_INET;
	session->serv_addr.sin_addr.s_addr = 
		                         inet_addr(session->my_inet_addr);
	session->serv_addr.sin_port        = htons(session->thePort);

	if ((session->theSocket = socket(AF_INET, SOCK_STREAM, 0)) < 0)
	{
		fprintf(stderr, "TCP EERROR : Could not open socket\n");
		return -1;
	}

	if (connect(session->theSocket,
               (struct sockaddr *) &session->serv_addr,
			    sizeof(session->serv_addr)) < 0)
	{
		fprintf(stderr, "TCP EERROR : Could not connect to server\n");
		close(session->theSocket);
		return -1;
	}

	session->connected++;

	if (txt_data)
	{
   		long dataLen1, dataLen2, i, l;

		/*--
			 Let's send the data to the 
			 other end of the socket
		--*/
		dataLen1 = charCount;
		i = htonl(dataLen1);

		/*-- send length of data --*/
		write(session->theSocket, (void *)&i, 4);

#ifdef WIN32
		/*-- convert ascii to ebcdic --*/
		convertAsciiToEbcdic(txt_data, dataLen1, 
			                 txt_data, dataLen1);
#endif
		/*-- send the data --*/
		txtPtr = txt_data;

		/*-- loop around until all data is sent --*/
		while (dataLen1)
		{
			i = write(session->theSocket, txtPtr, dataLen1);
			dataLen1 -= i;
			txtPtr += i;
		}

		/*--
			 Let's read the reply from the 
			 other end of the socket. This
			 will block until data is there
			 to read...
		--*/

		/*-- read length of reply --*/
		i = read(session->theSocket, (void *)&l, 4);

		dataLen1 = ntohl(l);
		dataLen2 = dataLen1;

		/*-- allocate memory for reply data --*/
		reply_data = calloc(dataLen1 + 1, 1);
		replyPtr = reply_data;

		/*-- get reply data, loop around until all read --*/
		while(dataLen1)
		{
			i = read(session->theSocket, replyPtr, dataLen1);
			dataLen1 -= i;
			replyPtr += dataLen1;
		}

		/*-- At this point, reply_data now 
			 contains the data sent from the
			 other end of the socket
		--*/
#ifdef WIN32
		/*-- convert ebcdic to ascii --*/
		convertEbcdicToAscii(reply_data, dataLen2, 
			                 reply_data, dataLen2);
#endif
		fprintf(stdout, 
				 "Message from server :\n%s\n", reply_data);
	    if ((fpOut = fopen(outFilename, "wb")) == NULL) {
 	        fprintf(stderr, "FILE ERROR : file 'outfile' not created\n");
            return -11;
		}	
	    else {
            charCount = fwrite(reply_data, 1, dataLen2, fpOut);
	        printf("%d chars written\n", charCount);
		    fclose(fpOut);
		}
		/*-- finished with reply data --*/
		free(reply_data);
	}

	close(session->theSocket);
	free(session);
	return 0;
}


char *show_address(struct in_addr *myaddr)
{
    if (myaddr)
    {
      static char buf[256];
      long host;

        host = ntohl(myaddr->s_addr);

        sprintf(buf,"%d.%d.%d.%d",
                (int)(0xff & (host >> 24)),
                (int)(0xff & (host >> 16)),
                (int)(0xff & (host >> 8)),
                (int)(0xff & host));
        return buf;
    }
	return 0;
}

#ifdef WIN32
/*------------------------------------------------------------------
Default CSSID 037 to ASCII tables
------------------------------------------------------------------*/
const unsigned char ASCII2EBCDIC[256]={
	/* 0000 */ 0x00,0x01,0x02,0x03,0x37,0x2d,0x2e,0x2f,
	/* 0008 */ 0x16,0x05,0x25,0x0b,0x0c,0x0d,0x0e,0x0f,
	/* 0010 */ 0x10,0x11,0x12,0x13,0xb6,0xb5,0x32,0x26,
	/* 0018 */ 0x18,0x19,0x3f,0x27,0x1c,0x1d,0x1e,0x1f,
	/* 0020 */ 0x40,0x5a,0x7f,0x7b,0x5b,0x6c,0x50,0x7d,
	/* 0028 */ 0x4d,0x5d,0x5c,0x4e,0x6b,0x60,0x4b,0x61,
	/* 0030 */ 0xf0,0xf1,0xf2,0xf3,0xf4,0xf5,0xf6,0xf7,
	/* 0038 */ 0xf8,0xf9,0x7a,0x5e,0x4c,0x7e,0x6e,0x6f,
	/* 0040 */ 0x7c,0xc1,0xc2,0xc3,0xc4,0xc5,0xc6,0xc7,
	/* 0048 */ 0xc8,0xc9,0xd1,0xd2,0xd3,0xd4,0xd5,0xd6,
	/* 0050 */ 0xd7,0xd8,0xd9,0xe2,0xe3,0xe4,0xe5,0xe6,
	/* 0058 */ 0xe7,0xe8,0xe9,0xba,0xe0,0xbb,0xb0,0x6d,
	/* 0060 */ 0x79,0x81,0x82,0x83,0x84,0x85,0x86,0x87,
	/* 0068 */ 0x88,0x89,0x91,0x92,0x93,0x94,0x95,0x96,
	/* 0070 */ 0x97,0x98,0x99,0xa2,0xa3,0xa4,0xa5,0xa6,
	/* 0078 */ 0xa7,0xa8,0xa9,0xc0,0x4f,0xd0,0xa1,0xff,
	/* 0080 */ 0x68,0xdc,0x51,0x42,0x43,0x44,0x47,0x48,
	/* 0090 */ 0x71,0x9c,0x9e,0xcb,0xcc,0xcd,0xdb,0xdd,
	/* 0098 */ 0xdf,0xec,0xfc,0x4a,0xb1,0xb2,0xbf,0x07,
	/* 00a0 */ 0x45,0x55,0xce,0xde,0x49,0x69,0x9a,0x9b,
	/* 00a8 */ 0xab,0xaf,0x5f,0xb8,0xb7,0xaa,0x8a,0x8b,
	/* 00b0 */ 0x2b,0x2c,0x09,0x21,0x28,0x65,0x62,0x64,
	/* 00b8 */ 0xb4,0x38,0x31,0x34,0x33,0x70,0x80,0x24,
	/* 00c0 */ 0x22,0x17,0x29,0x06,0x20,0x2a,0x46,0x66,
	/* 00c8 */ 0x1a,0x35,0x08,0x39,0x36,0x30,0x3a,0x9f,
	/* 00d0 */ 0x8c,0xac,0x72,0x73,0x74,0x0a,0x75,0x76,
	/* 00d8 */ 0x77,0x23,0x15,0x14,0x04,0x6a,0x78,0x3b,
	/* 00e0 */ 0xee,0x59,0xeb,0xed,0xcf,0xef,0xa0,0x8e,
	/* 00e8 */ 0xae,0xfe,0xfb,0xfd,0x8d,0xad,0xbc,0xbe,
	/* 00f0 */ 0xca,0x8f,0x1b,0xb9,0x3c,0x3d,0xe1,0x9d,
	/* 00f8 */ 0x90,0xbd,0xb3,0xda,0xfa,0xea,0x3e,0x41};
	
const unsigned char EBCDIC2ASCII[256]={
	/* 0100 */ 0x00,0x01,0x02,0x03,0xdc,0x09,0xc3,0x9f,
	/* 0108 */ 0xca,0xb2,0xd5,0x0b,0x0c,0x0d,0x0e,0x0f,
	/* 0110 */ 0x10,0x11,0x12,0x13,0xdb,0x0a,0x08,0xc1,
	/* 0118 */ 0x18,0x19,0xc8,0xf2,0x1c,0x1d,0x1e,0x1f,
	/* 0120 */ 0xc4,0xb3,0xc0,0xd9,0xbf,0x0a,0x17,0x1b,
	/* 0128 */ 0xb4,0xc2,0xc5,0xb0,0xb1,0x05,0x06,0x07,
	/* 0130 */ 0xcd,0xba,0x16,0xbc,0xbb,0xc9,0xcc,0x04,
	/* 0138 */ 0xb9,0xcb,0xce,0xdf,0xf4,0xf5,0xfe,0x1a,
	/* 0140 */ 0x20,0xff,0x83,0x84,0x85,0xa0,0xc6,0x86,
	/* 0148 */ 0x87,0xa4,0x9b,0x2e,0x3c,0x28,0x2b,0x7c,
	/* 0150 */ 0x26,0x82,0x88,0x89,0x8a,0xa1,0x8c,0x8b,
	/* 0158 */ 0x8d,0xe1,0x21,0x24,0x2a,0x29,0x3b,0xaa,
	/* 0160 */ 0x2d,0x2f,0xb6,0x8e,0xb7,0xb5,0xc7,0x8f,
	/* 0168 */ 0x80,0xa5,0xdd,0x2c,0x25,0x5f,0x3e,0x3f,
	/* 0170 */ 0xbd,0x90,0xd2,0xd3,0xd4,0xd6,0xd7,0xd8,
	/* 0178 */ 0xde,0x60,0x3a,0x23,0x40,0x27,0x3d,0x22,
	/* 0180 */ 0xbe,0x61,0x62,0x63,0x64,0x65,0x66,0x67,
	/* 0188 */ 0x68,0x69,0xae,0xaf,0xd0,0xec,0xe7,0xf1,
	/* 0190 */ 0xf8,0x6a,0x6b,0x6c,0x6d,0x6e,0x6f,0x70,
	/* 0198 */ 0x71,0x72,0xa6,0xa7,0x91,0xf7,0x92,0xcf,
	/* 01a0 */ 0xe6,0x7e,0x73,0x74,0x75,0x76,0x77,0x78,
	/* 01a8 */ 0x79,0x7a,0xad,0xa8,0xd1,0xed,0xe8,0xa9,
	/* 01b0 */ 0x5e,0x9c,0x9d,0xfa,0xb8,0x15,0x14,0xac,
	/* 01b8 */ 0xab,0xf3,0x5b,0x5d,0xee,0xf9,0xef,0x9e,
	/* 01c0 */ 0x7b,0x41,0x42,0x43,0x44,0x45,0x46,0x47,
	/* 01c8 */ 0x48,0x49,0xf0,0x93,0x94,0x95,0xa2,0xe4,
	/* 01d0 */ 0x7d,0x4a,0x4b,0x4c,0x4d,0x4e,0x4f,0x50,
	/* 01d8 */ 0x51,0x52,0xfb,0x96,0x81,0x97,0xa3,0x98,
	/* 01e0 */ 0x5c,0xf6,0x53,0x54,0x55,0x56,0x57,0x78,
	/* 01e8 */ 0x59,0x5a,0xfd,0xe2,0x99,0xe3,0xe0,0xe5,
	/* 01f0 */ 0x30,0x31,0x32,0x33,0x34,0x35,0x36,0x37,
	/* 01f8 */ 0x38,0x39,0xfc,0xea,0x9a,0xeb,0xe9,0x7f};
		
long convertAsciiToEbcdic(unsigned char *inBuffer,  long inLength, 
						  unsigned char *outBuffer, long outLength)
{
	long len = (inLength < outLength) ? inLength : outLength;
	long i;
	
	for (i = 0; i < len; i++)
	{
		outBuffer[(int)i] = ASCII2EBCDIC[inBuffer[(int)i]];
	}
	
	return 0;
}

long convertEbcdicToAscii(unsigned char *inBuffer,  long inLength, 
						  unsigned char *outBuffer, long outLength)
{
	long len = (inLength < outLength) ? inLength : outLength;
	long i;
	
	for (i = 0; i < len; i++)
	{
		outBuffer[(int)i] = EBCDIC2ASCII[inBuffer[(int)i]];
	}
	
	return 0;
}
#endif
